-- == Custom HUD Functions by TehRealSalt ==
-- Global script that contains extra functions for any HUD-related scripts.

-- USAGE: Main part of customhud is the ability to creating and overwriting existing HUD items, and adding support for other HUD modifications.

-- customhud.SetupItem(itemname, modname) lets you create a new HUD item. "itemname" is the identifier for a HUD item, while "modname" is the identifier for what it belongs to so it can be overwritten. Please run only once on script load.

-- Examples: customhud.SetupItem("bossmeter", "ktdbossmeter") registers the HUD item "bossmeter" to the mod "ktdbossmeter"
--			 (If bossmeter already exists, then the modname of bossmeter will be changed to ktdbossmeter anyways.)
--			 customhud.SetupItem("rings", "coolrings") overwrites the ring counter that's loaded (typically vanilla's), and registers it to the mod "coolrings".
--			 customhud.SetupItem("lives", "vanilla") changes the lives counter back to normal. ("vanilla" is the modname reserved for HUD items that exist in normal SRB2, and cannot be set to new HUD items.)

-- After a new HUD item is set up, it can be enabled or disabled with customhud.enable(itemname) and customhud.disable(itemname), respectively.
-- You can determine if your HUD item should be rendered by checking if customhud.CheckEnabled(itemname) is true and customhud.CheckType(itemname) is your modname.
-- Other people can also support your HUD replacements by checking for the same things.
-- Beyond that, customhud.FindItem can also be used to check if a HUD item already exists or not.

-- This also comes with some extra drawing functions.
-- customhud.CustomFontString(v, x, y, str, name, space, width, flags, align, scale, color) draws a string using a different font.
-- "v" is the drawer variable supplied by a HUD hook. "x" and "y" are the coordinates to draw at. "str" is the string to draw.
-- "name" is the font name prefix to use (must be between 1-5 characters long). Each lump should be stored in a WAD as "YYYYYxxx", with the Y's being the font name and the x's being the byte of the character.
-- "space" is the spacing to use when a character doesn't exist. Defaults to 4 if omitted.
-- "width" is the spacing between characters that do exist. Set to nil to use each individual patch's width instead.
-- "flags" are the flags used for each patch. "align" is how it should be aligned (can be "left", "center", or "right").
-- "scale" rescales the string if set. "color" is the SKINCOLOR_ constant to remap the text to.

-- Examples: customhud.CustomFontString(v, 8, 16, "Hello World!", "CRFNT") draws a string using SRB2's credits font
--			 customhud.CustomFontString(v, 16*FRACUNIT, 8*FRACUNIT, "World, Hello!", "LTFNT", nil, nil, 0, "left", 2*FRACUNIT) draws a rescaled string using SRB2's level title font
--			 customhud.CustomFontString(v, 16, 8, "Ayy, The World!", "COOLF", 8, 8, 0, "left", nil, SKINCOLOR_BLUE) draws a recolored, monospaced string using a custom font

-- NOTE: customhud.CustomFontString technically just draws patches for you. This means string-only flags will not work, patch-only flags will work, and green instead of white is remapped to the color.

-- customhud.CustomFontStringWidth(v, str, name, space, width, scale) returns the width of a custom font string. Each variable is identical to customhud.CustomFontString's.

-- customhud.CustomNum(v, x, y, str, name, space, width, padding, flags, align, scale, color) is similar to customhud.CustomFontString, except draws a font made of only numbers.
-- The font's lumps should be stored in a WAD as "YYYYYYYx", with the Y's being the font name and the x being the number (0-9).
-- padding is the number of zeroes to pad with.

-- customhud.CustomNumWidth(v, str, name, space, width, padding, scale) returns the width of a custom font number. Each variable is identical to customhud.CustomNum's.

-- Lastly, customhud.GetVersionNum() grabs the version number of this library, in table format. Usually, if you need to check this, you'll only need to check the first entry.

-- DO NOT EDIT THIS FILE! Don't add your custom HUD elements in here! Do it with the customhud.SetupItem function in another script instead.
-- If you want to add a feature to this, please share your changes with TehRealSalt so it can be put in the main file.

local VERSIONNUM = {1}

if (rawget(_G, "customhud")) then
	local oldnum = customhud.GetVersionNum()
	local numlength = max(#VERSIONNUM, #oldnum)
	
	local loadednum = ""
	local newnum = ""
	
	local oldvers = false
	
	for i = 1,numlength
		local num1 = oldnum[i]
		local num2 = VERSIONNUM[i]
		
		if (num1 == nil) num1 = 0 end
		if (num2 == nil) num2 = 0 end
		
		if (loadednum == "")
			loadednum = "v"..num1
		else
			loadednum = $1.."."..num1
		end
		
		if (newnum == "")
			newnum = "v"..num2
		else
			newnum = $1.."."..num2
		end
		
		if (num1 < num2)
			oldvers = true
		end
	end
	
	if (oldvers)
		print("An old version of customhud was detected. (Loaded version is "..loadednum..", newer version is "..newnum..") Please make sure all of the mods loaded are using the latest version.")
	end
	
	return
end
	
rawset(_G, "customhud", {})

function customhud.GetVersionNum()
	local tempnum = {} -- Make sure you cannot overwrite the version number, that would be dumb :V
	for i = 1,#VERSIONNUM
		tempnum[i] = VERSIONNUM[i]
	end
	return tempnum
end

local huditems = {
	stagetitle = {
		type = "vanilla", -- Current mod name for the hud item ("vanilla" is a reserved mod name for reverting a modded hud item back to its original state, cannot be set for custom hud items)
		enabled = true, -- Is this hud item supposed to be visible?
		orighuditem = true -- Should not EVER be true for hud items that are custom
	},
	textspectator = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	score = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	time = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	rings = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	lives = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	weaponrings = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	powerstones = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	nightslink = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	nightsdrill = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	nightsrings = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	nightsscore = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	nightstime = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	nightsrecords = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	rankings = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	coopemeralds = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	tokens = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	},
	tabemblems = {
		type = "vanilla",
		enabled = true,
		orighuditem = true
	}
}

local function warn(str)
	print("\131WARNING: \128"..str)
end

local function finditem(item)
	for k,_ in pairs(huditems)
		if (k == item) then
			return huditems[k]
		end
	end
	return false
end

function customhud.FindItem(item)
	if (finditem(item) != nil) then
		return true
	else
		return false
	end
end

function customhud.enable(item)
	local x = finditem(item)
	
	if not (x) return false end
	
	x.enabled = true
	if (x.orighuditem) then
		if (x.type == "vanilla") then
			hud.enable(item)
		else
			hud.disable(item)
		end
	end
	return true
end

function customhud.disable(item)
	local x = finditem(item)
	
	if not (x) return false end
	
	x.enabled = false
	if (x.orighuditem) then
		hud.disable(item)
	end
	return true
end

function customhud.CheckEnabled(item)
	local x = finditem(item)
	if not (x) return false end
	return x.enabled
end

function customhud.CheckType(item)
	local x = finditem(item)
	if not (x) return nil end
	return x.type
end

function customhud.SetupItem(item, t)
	if not (tostring(t))
		warn("Invalid type string \""..t.."\" in customhud.SetupItem")
	end
	
	local x = finditem(item)
	if (x)
		if (x.type != t) then
			if (t == "vanilla") and not (x.orighuditem) then
				warn("Type string \"vanilla\" is reserved for specific hud items in customhud.SetupItem")
				return false
			else
				x.type = t
				if (x.orighuditem)
					if (t == "vanilla")
						hud.enable(item)
					else
						hud.disable(item)
					end
				end
				return true
			end
		else
			return false
		end
	else
		huditems[item] = {
			type = t,
			enabled = true,
			orighuditem = false
		}
		return true
	end
end

local function pname(name, wb) return name..string.format("%03d", wb) end

function customhud.CustomFontStringWidth(v, str, name, space, width, scale)
	if not (type(str) == "string") then
		warn("No string given in customhud.CustomFontStringWidth")
		return
	end
	
	if not (type(name) == "string") then
		warn("No font given in customhud.CustomFontStringWidth")
		return
	end
	
	if (name:find(" ")) then
		warn("Font name \""..name.."\" cannot have spaces in customhud.CustomFontStringWidth")
		return
	end
	
	if (name:len() > 5) or (name:len() < 1) then
		warn("Incorrect font name length in customhud.CustomFontStringWidth")
		return
	end
	
	local strwidth = 0
	
	if (str == "") then return strwidth end
	
	if not (type(space) == "number") then
		space = 4
		if (type(scale) == "number") then
			space = $ * scale
		end
	end
	
	local nextx = x
	local capsoffset = 32
	
	for i = 1,str:len() do
		local nextbyte = str:byte(i,i)
		local patchn = pname(name,nextbyte)
		if (nextbyte >= 65 and nextbyte <= 90) and not v.patchExists(patchn)
			nextbyte = $ + capsoffset
			patchn = pname(name,nextbyte)
		elseif (nextbyte >= 97 and nextbyte <= 122) and not v.patchExists(patchn)
			nextbyte = $ - capsoffset
			patchn = pname(name,nextbyte)
		end
			
		if (v.patchExists(patchn)) then
			if (type(width) == "number") then
				strwidth = $ + width
			else
				strwidth = $ + v.cachePatch(patchn).width
			end
		else
			strwidth = $ + space
		end
	end
		
	if (type(scale) == "number") then
		strwidth = $ * scale
	end
	
	return strwidth
end

function customhud.CustomFontString(v, x, y, str, name, space, width, flags, align, scale, color)
	if not (type(str) == "string") then
		warn("No string given in customhud.CustomFontString")
		return
	end
	
	if not (type(name) == "string") then
		warn("No font given in customhud.CustomFontString")
		return
	end
	
	if (name:find(" ")) then
		warn("Font name \""..name.."\" cannot have spaces in customhud.CustomFontString")
		return
	end
	
	if (name:len() > 5) or (name:len() < 1) then
		warn("Incorrect font name length in customhud.CustomFontString")
		return
	end
	
	if (str == "") then return end -- I don't think blank strings would render anything in any case, so let's just silently return
	
	if not (type(space) == "number") then
		space = 4
		if (type(scale) == "number") then
			space = $ * scale
		end
	end
	
	local wc
	if (color) then
		wc = v.getColormap(-1, color)
	end
	
	local nextx = x
	local capsoffset = 32
	
	if (align == "right" or align == "center") then
		local strwidth = customhud.CustomFontStringWidth(v, str, name, space, width, align, scale)
		
		if (align == "right") then
			nextx = x - strwidth
		elseif (align == "center") then
			nextx = x - (strwidth / 2)
		end
	end
	
	for i = 1,str:len() do
		local nextbyte = str:byte(i,i)
		local patchn = pname(name,nextbyte)
		local patch
		
		if ((nextbyte >= 65 and nextbyte <= 90) and v.patchExists(pname(name, nextbyte + 32))) and not (v.patchExists(patchn))
			nextbyte = $ + capsoffset
			patchn = pname(name,nextbyte)
		elseif ((nextbyte >= 97 and nextbyte <= 122) and v.patchExists(pname(name, nextbyte - 32))) and not (v.patchExists(patchn))
			nextbyte = $ - capsoffset
			patchn = pname(name,nextbyte)
		end
		
		if (v.patchExists(patchn)) then
			patch = v.cachePatch(patchn)
			
			if (type(scale) == "number") then
				v.drawScaled(nextx, y, scale, patch, flags, wc)
			else
				v.draw(nextx, y, patch, flags, wc)
			end
		end
		
		if (patch and patch.valid) then -- If the character exists, move the next patch over to the right by...
			if (type(width) == "number") then
				if (type(scale) == "number") then
					nextx = $ + (width * scale)
				else
					nextx = $ + width
				end
			else
				if (type(scale) == "number") then
					nextx = $ + (patch.width * scale)
				else
					nextx = $ + patch.width
				end
			end
		else
			nextx = $ + space
		end
	end
end


function customhud.CustomNumWidth(v, num, name, width, pad, scale)
	if not (type(num) == "number") then
		warn("No number given in customhud.CustomNumWidth")
		return
	end
	
	if not (type(name) == "string") then
		warn("No font given in customhud.CustomNumWidth")
		return
	end
	
	if (name:find(" ")) then
		warn("Font name \""..name.."\" cannot have spaces in customhud.CustomNumWidth")
		return
	end
	
	if (name:len() > 7) or (name:len() < 1) then
		warn("Incorrect font name length in customhud.CustomNumWidth")
		return
	end
	
	local numwidth = 0
	
	num = tostring($1)
	if (tonumber(pad)) then
		num = string.format("%0"..pad.."d", $1)
	end
	
	if (num == "") then return numwidth end
	
	for i = 1,num:len() do
		local nextnum = num:sub(i,i)
		if (v.patchExists(name..nextnum)) then
			if (type(width) == "number") then
				numwidth = $ + width
			else
				numwidth = $ + v.cachePatch(name..nextnum).width
			end
		end
	end
		
	if (type(scale) == "number") then
		numwidth = $ * scale
	end
	
	return numwidth
end

function customhud.CustomNum(v, x, y, num, name, width, pad, flags, align, scale, color)
	if not (type(num) == "number") then
		warn("No number given in customhud.CustomNum")
		return
	end
	
	if not (type(name) == "string") then
		warn("No font given in customhud.CustomNum")
		return
	end
	
	if (name:find(" ")) then
		warn("Font name \""..name.."\" cannot have spaces in customhud.CustomNum")
		return
	end
	
	if (name:len() > 7) or (name:len() < 1) then
		warn("Incorrect font name length in customhud.CustomNum")
		return
	end
	
	local wc
	if (color) then
		wc = v.getColormap(-1, color)
	end
	
	local nextx = x
	
	if (align == "right" or align == "center") then
		local numwidth = customhud.CustomNumWidth(v, num, name, width, pad, scale)
		
		if (align == "right") then
			nextx = x - numwidth
		elseif (align == "center") then
			nextx = x - (numwidth / 2)
		end
	end
	
	num = tostring($1)
	if (tonumber(pad)) then
		num = string.format("%0"..pad.."d", $1)
	end
	
	for i = 1,num:len() do
		local nextnum = num:sub(i,i)
		local patch
		
		if (v.patchExists(name..nextnum)) then
			patch = v.cachePatch(name..nextnum)
			
			if (type(scale) == "number") then
				v.drawScaled(nextx, y, scale, patch, flags, wc)
			else
				v.draw(nextx, y, patch, flags, wc)
			end
		end
		
		if (patch and patch.valid) then -- If the character exists, move the next patch over to the right by...
			if (type(width) == "number") then
				if (type(scale) == "number") then
					nextx = $ + (width * scale)
				else
					nextx = $ + width
				end
			else
				if (type(scale) == "number") then
					nextx = $ + (patch.width * scale)
				else
					nextx = $ + patch.width
				end
			end
		else
			nextx = $ + width
		end
	end
end
